;;;; (net http server) -- a framework for making HTTP servers.

;;  Copyright 2001, Evan Prodromou <evan@prodromou.san-francisco.ca.us>.

;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; either version 2 of the License, or
;;  (at your option) any later version.

;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.

;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(define-module (net http server)
  :use-module (net http message)
  :use-module (oop goops)
  :use-module (oop goops describe)
  :use-module (ice-9 format)
  :use-module (ice-9 threads)
  :use-module (ice-9 regex))

(export <http-server>
        run
        handle-conn
        on-request
        on-not-implemented
        on-get
        on-options
        on-head
        on-post
        on-put
        on-delete
        on-trace
        on-connect
        on-unknown)
 
;;; The version of this module

(define *module-version* "0.3.1")
(define *server-name* (string-append "(net http server)/" *module-version*))

;;; A super-class for HTTP servers.

(define-class <http-server> (<object>)
  (port #:accessor port #:init-value 80 #:init-keyword #:port)
  (queue-size #:accessor queue-size #:init-value 5 #:init-keyword #:queue-size)
  (name #:accessor name #:init-value *server-name* #:init-keyword #:name))

(define-method (run (server <http-server>))
  (let ((sock (socket AF_INET SOCK_STREAM 0)))
    (bind sock AF_INET INADDR_ANY (port server))
    (listen sock (queue-size server))
    (let loop ()
      (select (list sock) '() '())
      (let* ((pair (accept sock))
             (prt (car pair))
             (conn (cdr pair)))
        (make-thread handle-conn server prt conn))
      (loop))
    (close-port sock)))

(define-method (handle-conn (server <http-server>) prt conn)
  (let* ((req (request-from-port prt))
	 (resp (on-request server req)))
    (to-port resp prt)))

(define-method (on-request (server <http-server>) (req <http-request>))
  (let ((m (meth req)))
    (cond
      ((string=? m "GET") (on-get server req))
      ((string=? m "OPTIONS") (on-options server req))
      ((string=? m "HEAD") (on-head server req))
      ((string=? m "POST") (on-post server req))
      ((string=? m "PUT") (on-put server req))
      ((string=? m "DELETE") (on-delete server req))
      ((string=? m "TRACE") (on-trace server req))
      ((string=? m "CONNECT") (on-connect server req))
      (else (on-unknown server req)))))

(define-method (on-not-implemented (server <http-server>) (req <http-request>))
  (let ((resp (make <http-response> #:code 501 #:status "Not Implemented"))
	(msg (string-append "Method '" (meth req) "' not implemented.\n")))
    (header-set! resp "Server" *server-name*)
    (header-set! resp "Content-Type" "text/plain")
    (header-set! resp "Content-Length" (string-length msg))
    (call-with-input-string
     msg
     (lambda (port)
       (set! (content resp) port)))
    resp))

(define-method (on-get (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

(define-method (on-options (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

(define-method (on-head (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

(define-method (on-post (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

(define-method (on-put (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

(define-method (on-delete (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

(define-method (on-trace (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

(define-method (on-connect (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

(define-method (on-unknown (server <http-server>) (req <http-request>))
  (on-not-implemented server req))

